//-----------------------------------------------------------------------------
// File: AILibAPI.h
//
// Desc: Extension API for AILib (used to enhance native classes' functionality)
//
// Copyright (c) 2000-2008 Black Element Software. All rights reserved
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// API EXTENSIONS
//-----------------------------------------------------------------------------

/** \addtogroup AIAPI */
/**@{*/
//! Solving AI weapons using
class WeaponManager
{
	// gatling handling
	float MIN_SHOOT_TIME;
	float MAX_SHOOT_TIME;
	float MIN_IDLE_TIME;
	float MAX_IDLE_TIME;

	const int FIRE_SINGLE = 0;
	const int FIRE_DOSES = 1;
	const int FIRE_REQUEST = 2;

	float Duration;	//! how long does current action last
	float Stamp;		//! when did it start
	bool Shooting;

	AIWeapon PrimaryWeapon; //! either gatling-like or walrus heavy weapon
	AIWeapon SecondaryWeapon; //! rocket-like for manta, possibly grenades for soldiers

	int PrimaryFireType;
	int SecondaryFireType;

	// some helpers
	int GetFireType( int weaponEnum );

	int RemainingAmmo()
	{
		int remain = 0;
		
		if(PrimaryWeapon)
		{
			remain += PrimaryWeapon.GetRemainingMagazineAmmo();
			remain += PrimaryWeapon.GetRemainingBackupAmmo();
		}
		if(SecondaryWeapon)
		{
			remain += SecondaryWeapon.GetRemainingMagazineAmmo();
			remain += SecondaryWeapon.GetRemainingBackupAmmo();
		}
		return remain;
	}

	void WeaponManager( Object parent )
	{
		int cnt = parent.GetWeaponCount();
		// strictly limit to 2 weapons
		switch( cnt )
		{
		case 0:
			PrimaryWeapon = NULL;
			SecondaryWeapon = NULL;
			break;

		case 1:
			PrimaryWeapon = parent.GetWeapon( 0 );
			SecondaryWeapon = NULL;
			break;

		case 2:
			PrimaryWeapon = parent.GetWeapon( 0 );
			SecondaryWeapon = parent.GetWeapon( 1 );

			//TEST: which one is primary in reality?
			if( SecondaryWeapon.GetCadence() > PrimaryWeapon.GetCadence() )
			{
				AIWeapon temp = PrimaryWeapon;
				PrimaryWeapon = SecondaryWeapon;
				SecondaryWeapon = temp;
			}
			break;

		default:
			Print( "Invalid number of weapons!" );
			PrimaryWeapon = NULL;
			SecondaryWeapon = NULL;
		}

		// init primary
		if( PrimaryWeapon )
		{
			PrimaryFireType = GetFireType( PrimaryWeapon.GetType() );
			PrimaryWeapon.EnableAutoFire( false );
		}

		// init secondary
		if( SecondaryWeapon )
		{
			SecondaryFireType = GetFireType( SecondaryWeapon.GetType() );
			SecondaryWeapon.EnableAutoFire( false );
		}

		// init times (primary weapon)
		// we assume secondary weapon is single shot ( :( )
		if( !PrimaryWeapon ) return;
		//float cadence = PrimaryWeapon.GetCadence();
		MIN_SHOOT_TIME = 400.0;
		MAX_SHOOT_TIME = 3000.0;
		MIN_IDLE_TIME = 100.0;
		MAX_IDLE_TIME = 600.0;
	}

	//! auto fire (primary weapon)
	void AutoFire()
	{
		if( PrimaryWeapon )
		{
			ditempentity = PrimaryWeapon.GetParent().GetEntity();
			ditempvale = g_Game.SearchArray( diunit, ditempentity );
			if ( ditempvale > 0 && ditempvale < 5 && !dimissilecanfire[ditempvale] )
				{
				if ( PrimaryWeapon.GetType() == wMantaMissile || PrimaryWeapon.GetType() == wMantaRocket )
					return;
				}
			float time = GetWorldTime();
			float dt = time - Stamp;

			switch( PrimaryFireType )
			{
			case FIRE_SINGLE: // solve single shot
				if( dt >= Duration && PrimaryWeapon.CanFire() )
				{
					// next shot
					PrimaryWeapon.EnableFire();

					Stamp = time;
					Duration = frand( MIN_IDLE_TIME, MAX_IDLE_TIME );
				}
				break;

			case FIRE_DOSES: // solve doses
				if( dt >= Duration )
				{
					if( Shooting ) // stop shooting
					{
						Shooting = false;
						Stamp = time;
						Duration = frand( MIN_IDLE_TIME, MAX_IDLE_TIME );
						PrimaryWeapon.EnableAutoFire( false );
					}
					else // start shooting
					{
						if( PrimaryWeapon.CanFire() )
						{
							Shooting = true;
							Stamp = time;
							Duration = frand( MIN_SHOOT_TIME, MAX_SHOOT_TIME );
							PrimaryWeapon.EnableAutoFire( true );
						}
					}
				}
				break;
			}
		}
	}

	//! reset the solver after target has got out of sight or something
	void Reset()
	{
		Duration = 0;
		Stamp = 0;
		if( PrimaryWeapon && Shooting ) PrimaryWeapon.EnableAutoFire( false );
		Shooting = false;
	}

	//! fire secondary weapon
	void FireSecondary()
	{
		if( SecondaryWeapon && SecondaryWeapon.CanFire() )
		{
			SecondaryWeapon.EnableFire();
		}
	}

	//! ask for time till reload will be needed
	float TimeToReload()
	{
		//TODO:
	}

	//! calculate aim vector against specified target
	vector CalculateAimVec( Object source, Object against, bool bPrimary )
	{
		AIWeapon w;
		if( bPrimary ) w = PrimaryWeapon;
		else w = SecondaryWeapon;
		if( !w ) return ZeroVec;

		vector tgPos = against.GetPosition();
		float tth = w.GetTimeToHit( tgPos );
		//TEST: skill
		InteractiveEntity ie = source.GetEntity();
		tth *= ie.GetSkill();
		vector vel = against.GetRealVelocity();
		vector realAim = vel * tth + tgPos;

		vector myPos = source.GetPosition();
		vector toTrg = realAim - myPos;
		VectorNormalize( toTrg );
		return toTrg;
	}
};

const int DOSE_WEAPONS[28] = { wWalrusLaser, wWalrusLaserGreen, wWalrusMachineGun, wWalrusGatling, wWalrusPlasma, wWalrusAA, wWalrusRepairGun,
wWalrusFlak, wMantaMachineGuns, wMantaGatlings, wMantaPlasmas, wMantaLasers, wMantaLaserGreen, wCharacterRifle,
wCharacterLaser, wCharacterPlasma, wDroidRifle, wDroidLaser, wDroidPlasma,
wTurretMachineGun, wTurretGatling, wTurretLightLaser, wTurretFlak, wTurretLightAA,
wTurretAA, wTurretPlasma, wTurretRepairGun,	wUECRifle };
const int SINGLE_WEAPONS[12] = { wWalrusLauncher, wWalrusGrenade, wWalrusShell, wWalrusHeat,
wWalrusHowitzer, wMantaMissile, wMantaRocket, wCharacterGrenade, wDroidGrenade,
wTurretGrenade,	wTurretShell, wTurretMissile };

int WeaponManager::GetFireType( int weaponEnum )
{
	for( int i = 0; i < sizeof( DOSE_WEAPONS ); i++ )
	{
		int w = DOSE_WEAPONS[i];
		if( weaponEnum == w ) return FIRE_DOSES;
	}

	for( i = 0; i < sizeof( SINGLE_WEAPONS ); i++ )
	{
		w = SINGLE_WEAPONS[i];
		if( weaponEnum == w ) return FIRE_SINGLE;
	}

	return FIRE_REQUEST;
}

class ScriptedCombatSolver;

//! Solving AI in combat
class ScriptedCombatSequence extends CombatSequence
{
	int PrefZone;
	WeaponManager WeaponSystem;

	// preferred distances
	float MAX_DISTANCE;
	float OPTIMAL_DISTANCE;
	float MIN_DISTANCE;
	Object Parent;
	ScriptedCombatSolver ParentSolver;

	//! retrieve combat sequence output
	// main method, must be overwritten by derived classes
	vector GetMovementTarget()
	{
	}

	//! event called when target changes during fight
	void OnTargetChange(Object lastTarget, Object newTarget)
	{
	}

	//! event called when target resets to NULL
	//TEMP: placeholder only, won't be ever called in this version - DO NOT OVERRIDE!
	void OnTargetReset()
	{
	}

	//! calculate preferred distances
	void UpdateDistances()
	{
		if( WeaponSystem && WeaponSystem.PrimaryWeapon )
		{
			int behavior = GetParent().GetCombatBehavior();
			MAX_DISTANCE = WeaponSystem.PrimaryWeapon.GetMaxShootRange();
			float minMod = 0.4;
			float optMod = 0.6;
			if( behavior == COMBAT_BERSERK )
			{
				// close up fight
				minMod = 0.2;
				optMod = 0.3;
			}
			else
			{
				if( behavior >= COMBAT_CAUTIOUS )
				{
					if( behavior >= COMBAT_PARANOID )
					{
						minMod = 0.6;
						optMod = 0.8;
					}
					else
					{
						minMod = 0.3;
						optMod = 0.7;
					}
				}
			}
			MIN_DISTANCE = MAX_DISTANCE * minMod;
			OPTIMAL_DISTANCE = MAX_DISTANCE * optMod;

			//TODO: apply spread modifier
		}
	}

	//! fire from all weapons
	void Fire()
	{
		Object parent = GetParent();
		if( parent == NULL ) return;

		int w = parent.GetWeaponCount();
		for( int i = 0; i < w; i++ )
		{
			parent.GetWeapon( i ).EnableFire();
		}
	}

	void EnableAutoFire( bool enable )
	{
		Object parent = GetParent();
		if( parent == NULL ) return;

		int w = parent.GetWeaponCount();
		for( int i = 0; i < w; i++ )
		{
			parent.GetWeapon( i ).EnableAutoFire( enable );
		}
	}

	//! called from solver
	bool CanDropTarget( Object target )
	{
		return true;
	}
	
	bool ProvidesCover( vector tgPos);
	vector GetRealCover(vector TgPos);
	bool RemoveCover();
	bool FindCover(vector myPos, float radius, Object enemy, Cover lastCover);

	void ~ScriptedCombatSequence()
	{
	}
};

//-----------------------------------------------------------------------------

//! CombatSolver base
class ScriptedCombatSolver extends CombatSolver
{
	// weapon system
	WeaponManager WeaponSystem;
	Object Parent;

	Cover MyCover;
	int CoverSlot;

	void ~ScriptedCombatSolver()
	{
		if(WeaponSystem)
		{
			delete WeaponSystem;
			WeaponSystem = NULL;
		}
		if( MyCover )
		{
			MyCover.UnlockCoverSlot( CoverSlot );
			MyCover = NULL;
		}
	}

	vector Process( Object target, int hz )
	{
	}

	void OnUpdateEquipment()
	{
		if(WeaponSystem)
		{
			delete WeaponSystem;
			WeaponSystem = NULL;
		}
	}

	//! event called when target changes during fight
	void OnTargetChanged( Object lastTarget, Object newTarget )
	{
	}

	//! event called when target resets to NULL
	void OnTargetReset( Object lastTarget )
	{
	}

	//! enable auto fire on all weapons
	void EnableAutoFire( bool enable )
	{
		Object parent = GetParent();
		if( parent == NULL ) return;

		int w = parent.GetWeaponCount();
		for( int i = 0; i < w; i++ )
		{
			parent.GetWeapon( i ).EnableAutoFire( enable );
		}
	}

	//! ask about the possibility to loose target
	bool CanDropTarget( Object target )
	{
		return true;
	}
};
/**@}*/
//-----------------------------------------------------------------------------
bool ScriptedCombatSequence::ProvidesCover( vector tgPos)
{
	if(ParentSolver.MyCover)
	{
		return ParentSolver.MyCover.ProvidesCover( Parent, tgPos, ParentSolver.CoverSlot );
	}
	return false;
}

//-----------------------------------------------------------------------------
vector ScriptedCombatSequence::GetRealCover(vector TgPos)
{
	return ParentSolver.MyCover.GetRealCover( ParentSolver.CoverSlot, TgPos );
}

//-----------------------------------------------------------------------------
bool ScriptedCombatSequence::RemoveCover()
{
	if( ParentSolver.MyCover )
	{
		ParentSolver.MyCover.UnlockCoverSlot( ParentSolver.CoverSlot );
		ParentSolver.MyCover = NULL;
		ParentSolver.CoverSlot = -1;
		return true;
	}
	return false;
}

//-----------------------------------------------------------------------------
bool ScriptedCombatSequence::FindCover(vector myPos, float radius, Object enemy, Cover lastCover)
{
	Cover newCover;
	int newCoverSlot
			 
	newCover = FindClosestCover( Parent, myPos, radius, enemy, newCoverSlot, lastCover );
	
	if(!newCover)
		return false;
	
	if(newCover != ParentSolver.MyCover || newCoverSlot != ParentSolver.CoverSlot)
	{
		newCover.LockCoverSlot(newCoverSlot, Parent);
		if(ParentSolver.MyCover)
		{
			ParentSolver.MyCover.UnlockCoverSlot(ParentSolver.CoverSlot);
		}
		ParentSolver.MyCover = newCover;
		ParentSolver.CoverSlot = newCoverSlot;
	}
	return true;
}

/** \addtogroup AIAPI */
/**@{*/
//! TargetSelector base
class ScriptedTargetSelector extends TargetSelector
{
	Object  LastTarget;                         // Posledni target
	InteractiveEntity   Ent;                    // parent
	bool    dp;                                 // debug print
	int 	Count;
	int 	Behaviorr[32];                    	// Behavior jednotek
	Object 	Objects[32];                     	// identicke ID
#ifdef DEVELOPER
	string  TSdp[32];                           // dp pro jednotky
	string  Name;                               // jmeno parenta
	string  TS;                                 // TS debug string
#endif

	//! before new pricing cycle occurs, last known target is supplied
	void OnBeforePrice( Object lastTarget )
	{
	    if( Ent != NULL )
	    {
			if ( LastTarget != lastTarget )
			{
				if ( Ent && Ent.IsCarrierUnit() )
					dp = true;
			}
			else
				dp = false;
		}
		else
			return;

		LastTarget = lastTarget;
		Count = 0;
	}

	//! main pricing function
	float PriceTarget( Object pTarget, float fDist, float fDistLimit, bool bIsAttacker, bool bIsSuggested, bool bIsInLoS )
	{
		return 0;
	}

	//! after pricing is done, selected target is provided
	void OnAfterPrice( Object selectedTarget )
	{
	    // Toto bude NULL pokud pujde o Drona
	    if ( Ent == NULL )
			return;

		if ( selectedTarget == NULL || Count == 0 )
			return;

#ifdef DEVELOPER
		if ( dp && TS_DEBUG )
			Print( Name );
#endif
		int behna = -1;

		for( int i = 0; i <= Count; i++ )
		{
			if ( selectedTarget == Objects[i] )
			{
				behna = Behaviorr[i];
#ifdef DEVELOPER
				if( dp && TS_DEBUG )
				{
					string target = TSdp[i] + " Behavior: ";

					switch( Behaviorr[i] )
					{
					case COMBAT_BERSERK:
						target += "BERSERK";
						break;
					case COMBAT_AGGRESSIVE:
						target += "AGRESSIV";
						break;
					case COMBAT_CAUTIOUS:
						target += "CAUTIOUS";
						break;
					case COMBAT_DEFENSIVE:
						target += "DEFENSIVE";
						break;
					case COMBAT_PARANOID:
						target += "PARANOIND";
						break;
					case COMBAT_CHICKEN:
						target += "CHICKEN";
						break;

					default:
						break;
					}
				}
#endif
			}

#ifdef DEVELOPER			
			if ( dp && TS_DEBUG )
			{
				Print( TSdp[i] );
				//g_Game.ShowHint( TSdp[i], 8, false );
			}
#endif
		}

#ifdef DEVELOPER
		if( dp && TS_DEBUG )
		{
			Print( target );
			g_Game.ShowHint( target, 12, false );
			Print("[TS] >>> --------------------------------------------------------------------------------------- ");
		}
#endif
		GetParent().SetCombatBehavior( behna );
		return;
	}
	
	void OnUpdateEquipment()
	{
	}
	
};
/**@}*/

/** \addtogroup MissionsAPI */
/**@{*/
//! Tactic base
class ScriptedTactic extends Tactic
{
	//! event called on object assign
	void OnAssignObject( ObjectHandlerBase object )
	{
	}

	//! event called on object removal
	void OnRemoveObject( ObjectHandlerBase object )
	{
	}

	//! main update method
	void Update()
	{
	}

	//! offer this tactic an object
	// should return true if tactic wants this object
	bool Offer( ObjectHandlerBase object )
	{
		// don't accept any object by default
		return false;
	}

	//! returns if can hold borrowed unit itself
	// should return true only if contains list of borrowed units
	bool WillHandleBorrowed()
	{
		// don't  handle borrowed unit
		return false;
	}

	//! ask this tactic if it still needs an object
	// if this returns false, the object in question might be automatically removed
	// by tactic manager
	bool Needs( ObjectHandlerBase object )
	{
		// keep all objects by default
		return true;
	}

	//! validate tactic
	// if false is returned, this instance will most likely be deleted
	bool Validate()
	{
		// if this is not overloaded, something is probably wrong
		return false;
	}

	//! should return estimate of needed units of given type
	int EstimateUnitNeed( int type )
	{
		return 0;
	}

	// look if any of handler's object is out of ammo
	//can be used in any tactic
	bool OutOfAmmo( ObjectHandlerBase object )
	{
		Object objects[32];
		int objectCount = object.BuildObjectList( objects);

		for( int x = 0; x < objectCount; x++ )
		{
			Object unit = objects[x];
			int weapCnt = unit.GetWeaponCount();
			if ( weapCnt <= 0 )
				return true;

			for ( int i = 0; i < weapCnt; i++)
			{
				AIWeapon weapon = unit.GetWeapon(i); 
				if ( weapon.GetRemainingMagazineAmmo() > 0 )
				{
					return false;
				}
				else
				{
					if ( weapon.GetRemainingBackupAmmo() > 0 )
						return false;
					else
						return true;
				}
			}
		}
		return false;
	}

	//  return weapon force of handler's object(squad)
	int GetForce( ObjectHandlerBase object )
	{
		int force = 0;
		Object objects[32];
		int objectCount = object.BuildObjectList( objects );

		for( int i = 0; i < objectCount; i++ )
		{
			Object unit = objects[i];
			int weapCnt = unit.GetWeaponCount();

			for( int j = 0; j < weapCnt; j++ )
			{
				AIWeapon weapon = unit.GetWeapon(j); 

				int ammo = weapon.GetAmmoType();
				if( ammo == AMMO_HOMING )
				{
					if( weapon.GetRemainingMagazineAmmo() < 2 ) continue;
					force += 5;
				}
				else
				{
					if( weapon.GetRemainingMagazineAmmo() < 50 ) continue;
					force += 2;
				}
			}
		}

		return force;
	}


	int GetRandomizedForce()
	{
		return 0;
	}
};
/**@}*/

/** \addtogroup StrategicAPI */
/**@{*/
//! Tactics manager base
class ScriptedTacticsManager extends TacticsManager
{
	//! ctor
	void ScriptedTacticsManager()
	{
		/*ManagedObjectCnt = 0;
		POICnt = 0;
		UnassignedObjectCnt = 0;
		TacticCnt = 0;*/
	}

	//! retrieve data from underlying c++ structure
	void UpdateData()
	{
		/*GetManagedObjects( ManagedObjects, ManagedObjectCnt );
		GetPOIs( POIs, POICnt );
		if( POICnt > 0 )
		{
		debug
		}
		GetUnassignedObjects( UnassignedObjects, UnassignedObjectCnt );
		GetTactics( Tactics, TacticCnt );*/
	}

	//! distribute objects among tactics
	void GenerateTactics()
	{
	}

	//! event called on object add
	void OnObjectAdd( ObjectHandlerBase object )
	{
	}

	//! event called on object removal
	void OnObjectRemove( ObjectHandlerBase object )
	{
	}
};

//-----------------------------------------------------------------------------

class StrategyLogger;
class ScriptedExpert extends Expert
{
	CCPlayer MyPlayer;
	StrategyLogger Logger;

	void ScriptedExpert()
	{
		Logger = new StrategyLogger;
	}

	void ~ScriptedExpert()
	{
		if( Logger )
			delete Logger;
	}

	void AnalyzeNeeds( Memory mem )
	{
		// make no specific needs by default
	}

	void AnalyzeProposals( Memory mem )
	{
		// make no proposals by default
	}

	void RunSolution( int entryID, int op1, int op2 )
	{
	}

	void OnControlLost()
	{
	}

	/*!
	Some helper methods.
	TODO: could make those native to spped'em up...
	*/
	//@{

	IslandModel Stockpile()
	{
	if ( !diexecutions )
	{
	Object ditempobjects[500];
	Object ditempobjectsb[500];
	Object ditempobject;
	Building build;
	Object diobj;
	ObjectHandlerBase diobjh;
	for ( ditempval = 0; ditempval < 10; ditempval++ )
		{
		if ( diwaypointposset[ditempval] )
			{
			diwaypointposset[ditempval] = false;
			if ( ditempval == 9 )
				diunit1 = didrone;
			else
				diunit1 = diunit[ditempval];
			if ( !ditempval && dicurrent && !dicurrent.IsCarrierUnit() && dicurrent.GetAssembly().IsWalrus() != dicurrent.GetAssembly().IsManta() )
				diunit1 = dicurrent;
			if ( diunit1 && !ditwprev )
				{
				if ( diunit1.IsCargo() )
					diunit1 = diunit1.GetCargoTransfer();
				diobj = diunit1.GetAI();
				if ( diobj )
					{
					diobjh = diobj.GetHandler();
					if ( !diwaypointposoverride[ditempval] || diunit1.GetAssembly().IsManta() )
						ditemppos = diwaypointpos[ditempval];
					else
						ditemppos = FindClosestReachable(diwaypointpos[ditempval], 5000, diobj);
					diunit1.PlanClear();
					diobjh.PlanAction(CreateMovementAction (ditemppos), false, false);
					diunit1.AutoPilot( true );
					diunit1.PlanEnable();
					}
				}
			}
		EntityBasic assembly = diPlayerUEC.GetAssembly(ditempval);
		if ( ditempval < 9 && dicarruec && assembly && assembly.GetEntity() && !diisdocked[ditempval] && !assembly.IsFullyDeployed() )
			{
			ditempvald = 1;
			diobj = assembly.GetEntity().GetAI();
			if ( diobj && diammoboxfix[ditempval] )
				{
				ditempvald = 0;
				for ( ditempvalb = 0; ditempvalb < diobj.GetWeaponCount(); ditempvalb++ )
					{
					if ( diobj.GetWeapon(ditempvalb).GetRemainingMagazineAmmo() < diobj.GetWeapon(ditempvalb).GetMagazineCapacity() )
						ditempvald = 1;
					}
				}
			for ( ditempvalb = 0; ditempvalb < assembly.GetEquipmentCount(); ditempvalb++ )
				{
				if ( !ditempvald || assembly.GetEquipment(ditempvalb) == wFlaresManta )
					assembly.SetEquipment( assembly.GetEquipment(ditempvalb) );
				}
			if ( !ditempvald )
				diammoboxfix[ditempval] = false;
			if ( ditempval < 5 && diobj )
				{
				if ( dimantaweapfix[ditempval] )
					{
					assembly.GetEntity().AddAmmo( -1000 );
					assembly.GetEntity().AddAmmo( dimantaweapfix[ditempval] );
					dimantaweapfix[ditempval] = 0;
					}
				else
					{
					if ( !assembly.GetEntity().GetAmmo() && diobj.GetWeaponCount() == 1 && diobj.GetWeapon(0).GetRemainingMagazineAmmo() )
						{
						dimantaweapfix[ditempval] = diobj.GetWeapon(0).GetRemainingMagazineAmmo();
						for ( ditempvalb = 0; ditempvalb < assembly.GetEquipmentCount(); ditempvalb++ )
							assembly.SetEquipment( assembly.GetEquipment(ditempvalb) );
						}
					else
						dimantaweapfix[ditempval] = 0;
					}
				}
			}
		}
	if ( dicarruec )
		{
		if ( dicarruec.GetLocalRepairZone( ECARRIER_ZONE_OFFENSE ).GetHitZoneComponent( 0 ) )
			{
			dicarruec.GetLocalRepairZone( ECARRIER_ZONE_OFFENSE ).GetHitZoneComponent( 0 ).SetPosition(dicarrweappos[0]);
			if ( dicurrent == dicarruec && ditacticalselectwidget && !dinohud && !diinmap )
				ShowWidget ( dihudframewidget, true );
			}
		for ( ditempval = 0; ditempval < 4; ditempval++ )
			{
			if ( dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( ditempval ) )
				dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( ditempval ).SetPosition(dicarrweappos[ditempval + 1]);
			}
		diobj = dicarruec.GetAI();
		for ( ditempval = 0; ditempval < diobj.GetWeaponCount(); ditempval++ )
			{
			ditempint = diobj.GetWeapon(ditempval).GetRemainingMagazineAmmo();
			ditempvalb = diobj.GetWeapon(ditempval).GetType();
			for ( ditempvalc = 1; ditempvalc < 16; ditempvalc++ )
				{
				if ( ditempvalb == dicarrweap[ditempvalc] )
					{
					if ( ditempint < dicarrweapammo[ditempvalc] )
						dicarrweapammoused[ditempvalc] = dicarrweapammoused[ditempvalc] + ( dicarrweapammo[ditempvalc] - ditempint );
					if ( dicarrweapammoused[ditempvalc] == 0 )
						dicarrweapammoused[ditempvalc] = diobj.GetWeapon(ditempval).GetMagazineCapacity() - ditempint;
					diunitpos = g_Game.GetCurrentCamera().GetPosition();
					if ( diunitpos[2] < 800 )
						diunitpos[2] = 0;
					ditemppos = dicarruec.GetLocalRepairZone( ECARRIER_ZONE_CONTROLTOWER ).GetCenter();
					ditemppos[2] = 0;
					ditempvald = VectorLength(ditemppos - diunitpos);
					ditempstring = "";
					if ( ditempvalb )
						ditempstring = " " + g_Game.ArticleName(ditempvalb) + " ";
					if ( dicarrweapicontextawidget[ditempvalc - 1 / 3 + 1 ] )
						SetWidgetTextEx(dicarrweapicontextawidget[ditempvalc - 1 / 3 + 1 ], 0, 0, ditempstring);
					if ( ditempvalb == wCarrierLeftLaser || ditempvalb == wCarrierRightLaser || ditempvalb == wCarrierRearLeftLaser || ditempvalb == wCarrierRearRightLaser )
						{
						ditempvale = 100 / diobj.GetWeapon(ditempval).GetMagazineCapacity();
						ditempvale *= diobj.GetWeapon(ditempval).GetMagazineCapacity() - dicarrweapammoused[ditempvalc];
						ditempvale += 0.5;
						if ( dicarrweapicontextbwidget[ditempvalc - 1 / 3 + 1 ] )
							SetWidgetTextEx(dicarrweapicontextbwidget[ditempvalc - 1 / 3 + 1 ], 0, 0, itoa(ditempvale) + "%");
						}
					else
						{
						if ( dicarrweapicontextbwidget[ditempvalc - 1 / 3 + 1 ] )
							SetWidgetTextEx(dicarrweapicontextbwidget[ditempvalc - 1 / 3 + 1 ], 0, 0, itoa(diobj.GetWeapon(ditempval).GetMagazineCapacity() - dicarrweapammoused[ditempvalc]) + " / " + itoa(diobj.GetWeapon(ditempval).GetMagazineCapacity()));
						}
 					if ( ditempint > diobj.GetWeapon(ditempval).GetMagazineCapacity() - dicarrweapammoused[ditempvalc] )
						{
						dicarrweapammofix[ditempvalc - 1 / 3] = true;
						ditempvale = 0;
						if ( ditempvalc > 0 && ditempvalc < 4 && ditempvald > 3450 && ditempvald < 3550 )
							ditempvale = 1;
						if ( ditempvalc > 3 && ditempvalc < 7 && ditempvald > 2800 && ditempvald < 3000 )
							ditempvale = 1;
						if ( ditempvalc > 6 && ditempvalc < 10 && ditempvald > 2600 && ditempvald < 2800 )
							ditempvale = 1;
						if ( ditempvalc > 9 && ditempvalc < 13 && ditempvald > 1200 && ditempvald < 1400 )
							ditempvale = 1;
						if ( ditempvalc > 12 && ditempvalc < 16 && ditempvald > 500 && ditempvald < 700 )
							ditempvale = 1;
						if ( ditempvale && !diunitpos[2] && dicurrent == dicarruec && !dicarrierselectwidget )
							{
							dicarruec.AddAmmo( diobj.GetWeapon(ditempval).GetMagazineCapacity() - dicarruec.GetAmmo() );
							dicarruec.AddAmmo( 0 - dicarrweapammoused[ditempvalc] );
							ditempint = dicarruec.GetAmmo();
							}
						else
							{
							if ( dicarrweapammoused[ditempvalc] >= diobj.GetWeapon(ditempval).GetMagazineCapacity() )
								{
								if ( ditempvalc > 0 && ditempvalc < 4 && dicarruec.GetLocalRepairZone( ECARRIER_ZONE_OFFENSE ).GetHitZoneComponent( 0 ) )
									dicarruec.GetLocalRepairZone( ECARRIER_ZONE_OFFENSE ).GetHitZoneComponent( 0 ).SetPosition("0 0 -100000");
								if ( ditempvalc > 9 && ditempvalc < 13 && dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( 0 ) )
									dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( 0 ).SetPosition("0 0 -100000");
								if ( ditempvalc > 6 && ditempvalc < 10 && dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( 1 ) )
									dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( 1 ).SetPosition("0 0 -100000");
								if ( ditempvalc > 3 && ditempvalc < 7 && dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( 2 ) )
									dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( 2 ).SetPosition("0 0 -100000");
								if ( ditempvalc > 12 && ditempvalc < 16 && dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( 3 ) )
									dicarruec.GetLocalRepairZone( ECARRIER_ZONE_DEFENSE ).GetHitZoneComponent( 3 ).SetPosition("0 0 -100000");
								}
							}
						}
					else
						dicarrweapammofix[ditempvalc - 1 / 3] = false;
					ditempvale = diobj.GetWeapon(ditempval).GetMagazineCapacity();
					if ( dicarrweapammoused[ditempvalc] > ditempvale )
						dicarrweapammoused[ditempvalc] = ditempvale;
					if ( dicarrweapammoused[ditempvalc] < ditempvale && dicarrweapammoused[ditempvalc] >= ditempvale * ( 1 - ( di_low_ammo_level / 100 ) ) )
						{
						if ( !dicarrweapammomsg[ditempvalc] )
							{
							g_Game.AddMessage( "CARRIER " + g_Game.ArticleName(ditempvalb) + " LOW AMMO", COLOR_YELLOW, 2 );
							g_Game.PlayGameSound( "game/radio/radio_unit_lowammo" );
							g_Game.PlayGameSound( "game/radio/radio_unit_lowammo" );
							g_Game.PlayGameSound( "game/radio/radio_unit_lowammo" );
							}
						dicarrweapammomsg[ditempvalc] = 1;
						}
					if ( di_low_ammo_level && dicarrweapammoused[ditempvalc] >= ditempvale && dicarrweapammomsg[ditempvalc] < 2 )
						{
						g_Game.AddMessage( "CARRIER " + g_Game.ArticleName(ditempvalb) + " NO AMMO", COLOR_RED, 2 );
						g_Game.PlayGameSound( "game/radio/radio_unit_noammo" );
						g_Game.PlayGameSound( "game/radio/radio_unit_noammo" );
						g_Game.PlayGameSound( "game/radio/radio_unit_noammo" );
						dicarrweapammomsg[ditempvalc] = 2;
						}
					if ( dicarrweapammoused[ditempvalc] < ditempvale * ( 1 - ( di_low_ammo_level / 100 ) ) )
						dicarrweapammomsg[ditempvalc] = 0;
					dicarrweapammo[ditempvalc] = ditempint;
					}
				}
			}
		for ( ditempval = 1; ditempval < 16; ditempval++ )
			{
			diPlayerUEC.GetCarrierStorage().RemoveItem( 13474 + ditempval, diPlayerUEC.GetCarrierStorage().GetItemCount( 13474 + ditempval ) );
			diPlayerUEC.GetCarrierStorage().AddItem( EGAMESTORAGE_COMMON, 13474 + ditempval, dicarrweapammoused[ditempval] );
			}
		}
	for ( ditempval = 1; ditempval < 5; ditempval++ )
		{
		ditempbuttonwidget = dimissileblockwidget[ditempval];
		ShowWidget ( ditempbuttonwidget, false );
		diunit1 = diunit[ditempval];
		dimissilecanfire[ditempval] = true;
		ditempwidget = FindWidget ( NULL, "CommandBarRoot.CommandBarUnitItem" + itoa(ditempval + 1) + ".missileblocktext" );
		ShowWidget ( ditempwidget, false );
		if ( diunit1 && diunit1.GetAssembly() )
			{
			ditempvale = 0;
			ditempvald = 0;
			ditempvalc = 0;
			diobj = diunit1.GetAI();
			if ( diobj )
				{
				for ( ditempvalb = 0; ditempvalb < diobj.GetWeaponCount(); ditempvalb++ )
					{
					if ( diobj.GetWeapon(ditempvalb).GetType() != wMantaMissile && diobj.GetWeapon(ditempvalb).GetType() != wMantaRocket )
						ditempvale += diobj.GetWeapon(ditempvalb).GetRemainingMagazineAmmo() + diobj.GetWeapon(ditempvalb).GetRemainingBackupAmmo();
					}
				if ( ditempvale && dimissileblock[ditempval] && diunit1.GetHealth() > 0.75 )
					dimissilecanfire[ditempval] = false;
				}
			if ( diunit1.HasEquip( wMantaMissile ) || diunit1.HasEquip( wMantaRocket ) )
				{
				if ( dilmb )
					ShowWidget ( ditempbuttonwidget, true );
				if ( dimissileblock[ditempval] )
					{
					ShowWidget ( ditempwidget, true );
					if ( dimissilecanfire[ditempval] && diobj )
						SetWidgetColor(ditempwidget, COLOR_YELLOW);
					else
						SetWidgetColor(ditempwidget, COLOR_RED);
					}
				}
			}
		}
	for ( ditempval = 1; ditempval < 9; ditempval++ )
		{
		ShowWidget( diammowidget[ditempval * 2 - 2], false );
		ShowWidget( diammowidget[ditempval * 2 - 1], false );
		diunit1 = diunit[ditempval];
		if ( di_ammo_level_bars && diunit1 && diunit1.GetAssembly() )
			{
			diobj = diunit1.GetAI();
			if ( diobj )
				{
				for ( ditempvalb = 0; ditempvalb < diobj.GetWeaponCount(); ditempvalb++ )
					{
					ditempint = diobj.GetWeapon(ditempvalb).GetType();
					for ( ditempvalc = 0; ditempvalc < sizeof(diitems); ditempvalc++ )
						{
						if ( diitems[ditempvalc] == ditempint )
							{
							ditempprogressbarwidget = NULL;
							if ( diequipslot[ditempvalc] == 1 )
								ditempprogressbarwidget = diammowidget[ditempval * 2 - 2];
							if ( diequipslot[ditempvalc] == 2 )
								ditempprogressbarwidget = diammowidget[ditempval * 2 - 1];
							if ( ditempprogressbarwidget )
								{
								ShowWidget( ditempprogressbarwidget, true );
								ditempvale = 0.28 * ( 100 / (diobj.GetWeapon(ditempvalb).GetMagazineCapacity() + diobj.GetWeapon(ditempvalb).GetBackupAmmoCapacity()) * ( diobj.GetWeapon(ditempvalb).GetRemainingMagazineAmmo() + diobj.GetWeapon(ditempvalb).GetRemainingBackupAmmo() ) )
								if ( ditempvale > 28 )
									ditempvale = 28;
								SetWidgetPos(ditempprogressbarwidget, 10, 110 - ( 30 * diequipslot[ditempvalc] ), 2, ditempvale );
								}
							}
						}
					}
				}
			}
		}
	for ( ditempval = 0; ditempval < 10; ditempval++ )
		{
		diunit1 = diattackmanta[ditempval];
		if ( diunit1 && dicarruec )
			{
			diobj = diunit1.GetAI();
			if ( diobj && diunit1.GetHealth() > 0 )
				{
				diunitpos = diunit1.GetPosition();
				ditempvalb = 0;
				diobjh = diobj.GetHandler();
				ditempvalc = diobj.GetWeaponCount();
				if ( ditempvalc > 0 )
					{
					AIWeapon diweap = diobj.GetWeapon(0);
					ditempvalb = diweap.GetRemainingMagazineAmmo();
					}
				if ( ditempvalb < 1 || ditempvalc > 1 && ditempvalb < diweap.GetMagazineCapacity() * 0.8 && ditempvalb > 50 || ditempvalb < diweap.GetMagazineCapacity() * 0.5 && ditempvalb > 50 || VectorLength( dicarrpos - "0 0 0" ) > diislrad[dicurrentisland2] + 40000 )
					{
					diattackmantaweapclear[ditempval] = diattackmantaweapclear[ditempval] + 1;
					if ( diattackmantaweapclear[ditempval] >= 20 )
						{
						if ( diunit1.GetAssembly().GetEquipmentCount() > 1 )
							{
							diunit1.DisableComplex();
							diunit1.ClearEquipment(diunit1.GetAssembly().GetEquipment(0));
							diunit1.EnableComplex();
							}
						else
							{
							diposz = 57.29 * atan2 ( dicarrpos[1], dicarrpos[0] );
							diposz += 180;
							diposx = dicarrpos[0] + ( cos(diposz * 0.017453) * 200000 );
							diposy = dicarrpos[1] + ( sin(diposz * 0.017453) * 200000 );
							diunit1.SetStance( STANCE_PASSIVE );
							diobjh.PlanAction(CreateMovementAction (Vector(diposx, diposy, 5000)), false, true);
							}
						}
					else
						diunit1.SetStance( STANCE_PASSIVE );
					}
				else
					{
					diattackmantaweapclear[ditempval] = 0;
					if ( VectorLength ( diunitpos - dicarrpos ) > 8000 )
						{
						diunit1.SetStance( STANCE_PASSIVE );
						diobjh.PlanAction(CreateMovementAction (dicarrpos), false, true);
						}
					else
						{
						if ( VectorLength ( diunitpos - Vector(dicarrpos[0], dicarrpos[1], diunitpos[2]) ) > 3000 )
							{
							diunit1.SetStance( STANCE_AGGRESSIVE );
							diobjh.CreateAttackAction( dicarruec.GetAI(), rand ( 0, 9 ) );
							if ( !dicarrattackwarn )
								{
								g_Game.AddMessage( "ENEMY LONG RANGE MANTAS ARE ATTACKING THE CARRIER!", COLOR_RED, 2 );
								dicarrattackwarn = true;
								}
							}
						else
							{
							if ( diunit1.GetStance() != STANCE_PASSIVE && VectorLength ( diunitpos - Vector(dicarrpos[0], dicarrpos[1], diunitpos[2]) ) < 1000 )
								{
								diunit1.SetStance( STANCE_PASSIVE );
								diunitpos[0] = diunitpos[0] + ( cos(diunit1.GetOrientation()[0] * 0.017453) * 5000 );
								diunitpos[1] = diunitpos[1] + ( sin(diunit1.GetOrientation()[0] * 0.017453) * 5000 );
								diobjh.PlanAction(CreateMovementAction (diunitpos), false, true);
								}
							}
						}
					}
				diunit1.EnableAI();
				diunit1.AutoPilot(true);
				diunit1.PlanEnable();
				}
			}
		}
	if ( didroneformation == 4 )
		ditempvale = -45;
	else
		{
		ditempvale = 0;
		if ( didroneformation != 5 )
			{
			if ( didronesout == 4 )
				ditempvale = -60;
			if ( didronesout == 3 )
				ditempvale = -50;
			if ( didronesout == 2 )
				ditempvale = -30;
			}
		}
	if ( didroneslaunch )
		dicarruec.LaunchDrones();
	if ( diddronespath )
		{
		vector didronetargetpos[4];
		didronesready = 0;
		for ( ditempval = 0; ditempval < 4; ditempval++ )
			{
			diunit1 = diddrone[ditempval];
			if ( diunit1 && dicarruec && !direset && !diresetprev )
				{
				diunitpos = diunit1.GetPosition();
				if ( didronesdocking )
					{
					if ( ditempval == 0 )
						ditempvale = -32;
					if ( ditempval == 1 )
						ditempvale = -10.5;
					if ( ditempval == 2 )
						ditempvale = 10.5;
					if ( ditempval == 3 )
						ditempvale = 32;
					didronetargetpos[ditempval] = Vector ( dicarrpos[0] + ( cos(dicarrori[0] + 180 - ditempvale * 0.017453) * 5500 ), dicarrpos[1] + ( sin(dicarrori[0] + 180 - ditempvale * 0.017453) * 5500 ), 210 );
					if ( VectorLength ( diunitpos - Vector(didronetargetpos[ditempval][0], didronetargetpos[ditempval][1], diunitpos[2] ) ) < 350 )
						{
						didronesready++;
						diunit1.Wait(1, 0);
						}
					}
				else
					{
					if ( didroneformation == 6 )
						ditempvald = 57.29 * atan2 ( dicarrpos[1] - dinearestwalruspos[1], dicarrpos[0] - dinearestwalruspos[0] );
					else
						ditempvald = dicarrori[0];
					diunit1.PlanClear();
					ditempvalb = 90 * didroneformation;
					didronetargetpos[ditempval] = Vector ( dicarrpos[0] + ( cos(ditempvald + ditempvale + ditempvalb * 0.017453) * 3500 ), dicarrpos[1] + ( sin(ditempvald + ditempvale + ditempvalb * 0.017453) * 3500 ), 0 );
					didroneinpos[ditempval] = false;
					if ( VectorLength ( diunitpos - Vector(didronetargetpos[ditempval][0], didronetargetpos[ditempval][1], diunitpos[2] ) ) < 350 )
						{
						didroneinpos[ditempval] = true;
						diunit1.Wait(1, 0);
						}
					}
				diobj = diunit1.GetAI();
				didronestuck[ditempval] = false;
				if ( diobj && !diunit1.GetStance() )
					{
					if ( !didronedockstart && !didroneundockstart && !dicarrmoving2 && VectorLength ( diunitpos - Vector(didronetargetpos[ditempval][0], didronetargetpos[ditempval][1], diunitpos[2] ) ) > 700 && VectorLength ( diunitpos - diddroneposprev[ditempval] ) < 20 )
						didronestuck[ditempval] = true;
					diobjh = diobj.GetHandler();
					diobjh.PlanAction(CreateIMovementAction (didronetargetpos[ditempval]) , false, false);
					diunit1.PlanEnable();
					diddroneposprev[ditempval] = diunitpos;
					}
				if ( didroneformation == 4 || didroneformation == 5 )
					ditempvale += 90;
				else
					{
					if ( didronesout == 4 )
						ditempvale += 40;
					if ( didronesout == 3 )
						ditempvale += 50;
					if ( didronesout == 2 )
						ditempvale += 60;
					}
				}
			}
		for ( ditempval = 0; ditempval < 4; ditempval++ )
			{
			for ( ditempvalb = 0; ditempvalb < 4; ditempvalb++ )
				{
				diunit1 = diddrone[ditempval];
				diunit2 = diddrone[ditempvalb];
				if ( ditempval != ditempvalb && diunit1 && diunit2 && VectorLength( diunit1.GetPosition() - didronetargetpos[ditempvalb] ) < VectorLength( diunit2.GetPosition() - didronetargetpos[ditempvalb] ) - 350 && VectorLength( diunit1.GetPosition() - didronetargetpos[ditempvalb] ) < VectorLength( diunit1.GetPosition() - didronetargetpos[ditempval] ) - 350 )
					{
					diddrone[ditempval] = diddrone[ditempvalb];
					diddrone[ditempvalb] = diunit1;
					}
				}
			}
		diddronespath = false;
		}
	if ( diemptymantacheck )
		diemptymantacheck++;
	if ( diemptymantacheck > 5 )
		diemptymantacheck = 0;
	if ( dicheckspawnstations && diinmap )
		{
		dicheckspawnstations--;
		for ( ditempval = 0; ditempval < dispawnstationmax; ditempval++ )
			{
			ditempbuilding = dispawnstation[ditempval];
			if ( !dicheckspawnstations &&ditempbuilding && ditempbuilding.GetBuildingType() == BuildingGarageWalrus && !ditempbuilding.GetHealth() )
				{
				ditemppos = ditempbuilding.GetPosition();
				ditempbuilding.SetPosition( "100000 100000 -100000" );
				ditempbuilding.Destroy();
				ditempbuilding.SetPosition( ditemppos );
				dispawnstation[ditempval] = NULL;
				}
			}
		}
	if ( dibuildingschecked < 2 || didroneslaunch || didronesdocked < didronesdockedprev || diemptymantacheck == 5 )
		{
		didronesdockedprev = didronesdocked;
		diislandunitcontrol = true;
		dibuildingschecked++;
		dispawnstationmax = 0;
		if ( dicarruec && dicarruec.GetCarrierTelemetry() < 90000 && !g_Game.GetCurrentIsland().GetPlayer().GetID() )
			g_Game.GetCurrentIsland().UniPostClear();
		for ( ditempvalc = 0; ditempvalc < 20; ditempvalc++ )
			dispawnstation[ditempvalc] = NULL;
		dient = GetNextEntity(NULL, Type("Building"));
		while (dient)
			{
			build = dient;
			dient = GetNextEntity(dient, Type("Building"));
			ditempvalc = build.GetBuildingType();
			if ( ditempvalc == BuildingGarageManta && build.GetHealth() || ditempvalc == BuildingGarageWalrus && build.GetHealth() )
				{
				dispawnstation[dispawnstationmax] = build;
				dispawnstationmax++;
				}
			}
		ditempvale = 0;
		dicontainer = FindNamedEntity( g_Game.CurrentWorld, "Container_Parent" );
		ditempentity = FindNamedEntity( g_Game.CurrentWorld, "MantaReal" );
		if ( ditempentity && dibuildingschecked == 1 && !digm )
			ditempentity.DisableComplex();
		dircactivator = FindNamedEntity( g_Game.CurrentWorld, "RC_Activator" );
		if ( dircactivator && !g_GM && dicurrenttime < 5 )
			dircactivator.EnableUse();
		for ( ditempval = 0; ditempval < g_Game.GetNumIslands(); ditempval++ )
			{
			ditempisl = g_Game.GetIsland(ditempval);
			ditemppos = ditempisl.GetStrategyPos();
			ditempvalb = VectorLength( "0 0 0" - ditemppos );
			if ( ditempvalb > ditempvale )
				{
				ditempvale = ditempvalb;
				ditempvalc = ditempval;
				}
			}
		for ( ditempval = 0; ditempval < 33; ditempval++ )
			{
			if ( ditempvalc > 32 )
				ditempvalc = 0;
			dircbpa[ditempvalc] = dircbparef[ditempval];
			dircbpb[ditempvalc] = dircbpbref[ditempval];
			ditempvalc++;
			}
		for ( ditempval = 1; ditempval < 9; ditempval++ )
			{
			diunit[ditempval] = diPlayerUEC.GetAssembly(ditempval).GetEntity();
			diunit1 = diunit[ditempval];
			if ( diunit1 )
				{
				if ( diunit1.GetHealth() <= 0 )
					{
					diunit1.Destroy();
					diunit[ditempval] = NULL;
					}
				}
			}
		for ( ditempvalb = 0; ditempvalb < 2; ditempvalb++ )
			{
			if ( !diislhasturrets )
				{
				for ( ditempvalc = 0; ditempvalc < 150; ditempvalc++ )
					dineutturret[ditempvalc] = NULL;
				}
			if ( ditempvalb && dicarruec )
				ditempval = g_Game.QueryUnitsBySphere( dicarruec.GetPosition(), 10000, ditempobjects, 500 );
			else
				ditempval = g_Game.QueryUnitsBySphere( "0 0 0", 1000000, ditempobjects, 500 );
			for ( ditempval = 0; ditempval < 500; ditempval++ )
				{
				ditempobject = ditempobjects[ditempval];
				if ( ditempobject )
					{
					ditempentity = ditempobject.GetEntity();
					if ( IsInherited( ditempentity,Type("Manta") ) && !ditempentity.IsComplexlyDisabled() && !ditempentity.IsCarrierUnit() && dicurrent && ditempentity != dicurrent.GetCargoTransfer() && strstr (ditempentity.GetName(), "Myrik" ) == -1 )
						{
						if ( ditempentity.GetPosition()[2] > 19000 )
							{
							ditempvale = 0;
							for ( ditempvalc = 0; ditempvalc < 10; ditempvalc++ )
								{
								if ( diattackmanta[ditempvalc] == NULL && !ditempvale )
									{
									diattackmanta[ditempvalc] = ditempentity;
									ditempvale = 1;
									}
								}
							}
						ditempvale = 0;
						for ( ditempvalc = 0; ditempvalc < 20; ditempvalc++ )
							{
							diunit2 = dispawnstation[ditempvalc];
							if ( diunit2 )
								{
								ditemppos = diunit2.GetPosition();
								diunitpos = ditempentity.GetPosition();
								ditemppos[2] = diunitpos[2];
								if ( VectorLength( diunitpos - ditemppos ) < 1 )
									ditempvale = 1;
								}
							}
						if ( !ditempvale )
							{
							ditempentity.DisableComplex();
							ditempentity.EnableComplex();
							if ( diemptymantacheck == 5 && !ditempentity.GetAssembly().GetEquipmentCount() )
								{
								ditempentity.SetPosition( "0 0 -1000" );
								ditempentity.Destroy();
								}
							}
						else
							diemptymantacheck = 1;
						if ( ditempentity != g_Game.GetCurrentUnit() )
							{
							ditempentity.AutoPilot( true );
							ditempentity.PlanEnable();
							}
						else
							ditempentity.AutoPilot( false );
						}
					if ( IsInherited( ditempentity,Type("Walrus") ) && !ditempentity.IsComplexlyDisabled() && !ditempentity.IsCarrierUnit() && strstr (ditempentity.GetName(), "Myrik" ) == -1  )
						{
						ditempentity.DisableComplex();
						if ( digm || ditempentity.GetAssembly().GetEquipmentCount() )
							ditempentity.EnableComplex();
						if ( ditempentity != g_Game.GetCurrentUnit() )
							{
							ditempentity.AutoPilot( true );
							ditempentity.PlanEnable();
							}
						else
							ditempentity.AutoPilot( false );
						}
					if ( ditempentity != didrone && IsInherited( ditempentity,Type("Drone") ) && ditempentity.GetPlayer() == diPlayerUEC && ditempentity.GetVObject() && ditempentity.GetHealth() )
						{
						if ( vtoa( ditempentity.GetVObject() ) == "obj/vehicles/uav/uav.xob" )
							{
							ditempvalb++;
							didrone = ditempentity;
							diaddsatelliteobject( ditempentity, "Icon40UAV", false, true );
							if ( dibuildingschecked == 1 && !diinmap )
								{
								g_Game.ShowMap();
								diselectprevnextframe = 4;
								}
							didroneheight = ditempentity.GetPosition()[2];
							if ( didrone.GetHealth() == 1 && didrone.GetFuel() >= 0.99 )
								{
								didrone.GetAssembly().SetHealth( didronehealthb / 100 );
								didrone.SetFuel( didronefuelb / 100 );
								}
							}
						else
							{
							if ( g_Game.SearchArray(diddrone, ditempentity) == -1 )
								{
								diaddsatelliteobject( ditempentity, "HudRadarUnit", false, true );
								for ( ditempvalc = 0; ditempvalc < 4; ditempvalc++ )
									{
									if ( diddrone[ditempvalc] == NULL )
										{
										diddrone[ditempvalc] = ditempentity;
										if ( didroneslaunch )
											{
											diunit1 = ditempentity;
											if ( ditempvalc == 0 )
												ditempvald = -3;
											if ( ditempvalc == 1 )
												ditempvald = -1;
											if ( ditempvalc == 2 )
												ditempvald = 1;
											if ( ditempvalc == 3 )
												ditempvald = 3;
											diunit1.SetPosition ( Vector ( dicarrpos[0] + ( cos(dicarrori[0] + 180 - ditempvald * 0.017453) * 2530 ), dicarrpos[1] + ( sin(dicarrori[0] + 180 - ditempvald * 0.017453) * 2530 ), 210 ) );
											}
										ditempvalc = 5;
										}
									}
								}
							}
						}
					if ( IsInherited( ditempentity,Type("Turret") )
						{
						Turret ditempturret = ditempentity;
						if ( !ditempturret.GetAssembly().GetEquipmentCount() )
							{
							ditempentity.SetPosition( "5 5 -5000" );
							ditempentity.Destroy();
							}
						else
							{
							ditempint = ditempturret.GetAttackType();
							if ( ditempint == 0 )
								ditempstring = "HudRadarTurretAA";
							if ( ditempint == 1 )
								ditempstring = "HudRadarTurretAT";
							if ( ditempint == 2 )
								{
								if ( ditempturret.GetAssembly().GetEquipment( 0 ) == 1309 )
									ditempstring = "HudRadarTurretMissile";
								else
									ditempstring = "HudRadarTurretAT";
								}
							diaddsatelliteobject( ditempentity, ditempstring, true );
							if ( !ditempentity.GetPlayer().GetID() )
								{
								diislhasturrets = 1;
								for ( ditempvalc = 0; ditempvalc < 150; ditempvalc++ )
									{
									if ( !dineutturret[ditempvalc] )
										{
										dineutturret[ditempvalc] = ditempentity;
										ditempvalc = 150;
										}
									}
								}
							}
						}
					if ( IsInherited( ditempentity,Type("Building") )
						{
						build = ditempentity;
						ditempvalc = build.GetBuildingType();
						if ( ditempvalc == BuildingPier || ditempvalc == BuildingFactory || ditempvalc == BuildingRefinery || ditempvalc == BuildingHeadQuarters || ditempvalc == BuildingUniPostFirewall || ditempvalc == BuildingGenericFirewall )
							{
							if ( di_protect_useful_buildings )
								build.SetInvincibility(true);
							else
								build.SetInvincibility(false);
							}
						if ( ditempvalc == BuildingGenericPump && build.WasDiscovered() )
							direfueldiscovered = true;
						if ( ditempvalc == BuildingGenericHealth && build.WasDiscovered() )
							direpairdiscovered = true;
						if ( ditempvalc == BuildingGenericRearm && build.WasDiscovered() )
							direarmdiscovered = true;
						if ( g_Game.SearchArray(dibuilding, build) == -1 )
							{
							for ( ditempvale = 0; ditempvale < 30; ditempvale++ )
								{
								if ( dibuilding[ditempvale] == NULL )
									{
									if ( ditempvalc == BuildingGenericFirewall )
										ditempvalc = BuildingUniPostFirewall;
									if ( ditempvalc == BuildingFactory || ditempvalc == BuildingRefinery || ditempvalc == BuildingHeadQuarters || ditempvalc == BuildingCommandCenter || ditempvalc == BuildingUniPostFirewall || ditempvalc == BuildingUniPostScrambler || ditempvalc == BuildingGenericPump ||  ditempvalc == BuildingGenericRearm ||  ditempvalc == BuildingGenericHealth )
										{
										dibuilding[ditempvale] = build;
										ditempvale = 0;
										if ( ditempvalc > 8 )
											ditempvale = 1;
										if ( ditempvalc > 100 )
											diaddsatelliteobject( ditempentity, disatelliteicon[ditempvalc - 100], true, ditempvale );
										else
											diaddsatelliteobject( ditempentity, disatelliteicon[ditempvalc], true, ditempvale );
										if ( ditempvalc == BuildingCommandCenter && diislstatus[dicurrentisland2] == 2 && build.WasDiscovered() )
											diforcerc = 11;
										break;
										}
									}
								}
							}
						if ( ditempvalc == BuildingPier && g_Game.SearchArray(dipier, build) == -1 )
							{
							for ( ditempvale = 1; ditempvale < 5; ditempvale++ )
								{
								if ( dipier[ditempvale] == NULL )
									{
									dipier[ditempvale] = build;
									diaddsatelliteobject( ditempentity, "HudRadarMolo", true );
									break;
									}
								}
							}
						if ( ditempvalc == BuildingGenericPump && g_Game.SearchArray(dipump, build) == -1 )
							{
							for ( ditempvale = 1; ditempvale < 7; ditempvale++ )
								{
								if ( dipump[ditempvale] == NULL )
									{
									dipump[ditempvale] = build;
									break;
									}
								}
							}
						if ( ditempvalc == BuildingRC )
							{
							diresearch = build;
							diaddsatelliteobject( ditempentity, "HudRadarResearchCenter", true, true );
							if ( diislstatus[dicurrentisland2] == 2 && diresearch.WasDiscovered() )
								diforcerc = 11;
							}
						if ( ditempvalc == BuildingGarageManta || ditempvalc == BuildingGarageWalrus || ditempvalc == BuildingRadar || ditempvalc == BuildingUniPostBooster )
							{
							if ( ditempvalc > 100 )
								ditempvalc -= 100;
							diaddsatelliteobject( ditempentity, disatelliteicon[ditempvalc], true, true );
							}
						}
					if ( IsInherited( ditempentity,Type("Character") )
						{
						ditemppos = ditempentity.GetPosition();
						if ( ditemppos[2] > 18000 )
							{
							ditempvalc = 10000000;
							ditempvald = 0;
							ditempentity.DisableComplex();
							ditempentity.SetPosition( Vector(ditemppos[0], ditemppos[1], GetTerrainHeight(ditemppos) + 50) );
							ditempentity.EnableComplex();
							if ( ditempentity == dicurrent && dicarruec )
								{
								dicarruec.SelectForPlayer();
								ditempentity.SelectForPlayer();
								}
							for ( ditempvale = 5; ditempvale < 9; ditempvale++ )
								{
								diunit1 = diunit[ditempvale];
								if ( diunit1 && diPlayerUEC.GetCarrierStorage().GetItemCount(13461 + ditempvale) && VectorLength(Vector(ditemppos[0], ditemppos[1], 0) - Vector(diunit1.GetPosition()[0], diunit1.GetPosition()[1], 0)) < ditempvalc )
									{
									ditempvalc = VectorLength(Vector(ditemppos[0], ditemppos[1], 0) - Vector(diunit1.GetPosition()[0], diunit1.GetPosition()[1], 0));
									ditempvald = ditempvale;
									}
								}
							if ( ditempvald )
								{
								diunit1 = diunit[ditempvald];
								diPlayerUEC.GetCarrierStorage().RemoveItem(13461 + ditempvald, 1);
								ditempvale = ditempvald * 4 - 19;
								for ( ditempvald = ditempvale; ditempvald < ditempvale + 4; ditempvald++ )
									{
									if ( !didroid[ditempvald] )
										{
										didroid[ditempvald] = ditempentity;
										if ( VectorLength(diunit1.GetPosition() - ditempentity.GetPosition()) < 3750 )
											didroidleader[ditempvald] = diunit1;
										break;
										}
									}
								}
							else
								{
								for ( ditempvald = 17; ditempvald < 33; ditempvald++ )
									{
									if ( !didroid[ditempvald] )
										{
										didroid[ditempvald] = ditempentity;
										break;
										}
									}
								}
							}
						}
					}
				}
			ditempvale = 0;
			for ( ditempval = 0; ditempval < 30; ditempval++ )
				{
				build = dibuilding[ditempval];
				if ( build && build.GetBuildingType() == BuildingCommandCenter )
					ditempvale = 1;
				}
			if ( !ditempvale )
				{
				build = g_Game.GetCommandCenter();
				for ( ditempval = 0; ditempval < 30; ditempval++ )
					{
					if ( !dibuilding[ditempval] )
						{
						dibuilding[ditempval] = build;
						break;
						}
					}
				}
			for ( ditempval = 0; ditempval < 30; ditempval++ )
				{
				ditempbuilding = dibuilding[ditempval];
				if ( ditempbuilding && ditempbuilding.GetBuildingType() == BuildingCommandCenter )
					{
					dicc = ditempbuilding;
					diccpos = dicc.GetPosition();
					if ( ditempbuilding.GetPlayer() == g_Game.GetPlayer(0) )
						{
						ditempbuilding.SetPlayer(diPlayerAPA);
						for ( ditempvale = 0; ditempvale < 150; ditempvale++ )
							{
							ditempentity = dineutturret[ditempvale];
							if ( ditempentity )
								{
								ditempentity.ForcePlayer(diPlayerAPA);
								ditempentity.DisableComplex();
								ditempentity.EnableComplex();
								}
							}
						ditempbuilding.SetPlayer(g_Game.GetPlayer(0));
						for ( ditempvale = 0; ditempvale < 150; ditempvale++ )
							{
							ditempentity = dineutturret[ditempvale];
							if ( ditempentity )
								ditempentity.ForcePlayer(g_Game.GetPlayer(0));
							}
						}
					}
				}
			}
		}
	didroneslaunch = false;
	disatelliterefresh++;
	if ( di_tactical_scan_limit && dicarrscancheck && !direset && !diresetprev || di_tactical_scan_limit && !di_scan_update_rate && !direset && !diresetprev || disatelliterefresh > 15 )
		{
		dinearestwalruspos = "0 0 0";
		if ( didrone && didrone.GetAI() )
			ditempentity = didrone;
		else
			ditempentity = dicarruec;
		if ( !ditempentity )
			ditempentity = dicurrent;
		if ( ditempentity && ditempentity.GetAI() && ditacticalrootwidget[0] && dicurrent )
			{
			int diobjectdist[500];
			float diobjectdistb[500];
			int dimaxresults = discreensizey / 13;
			if ( dimaxresults > di_tactical_scan_limit * 5 )
				{
				if ( dicurrentunit || dicurrent.GetAssembly().IsManta() || dicurrent.GetAssembly().IsWalrus() )
					dimaxresults = di_tactical_scan_limit * 5;
				}
			if ( dimaxresults > 100 )
				dimaxresults = 100;
			for ( ditempvale = 0; ditempvale < 500; ditempvale++ )
				{
				diobjectdist[ditempvale] = 0;
				diobjectdistb[ditempvale] = 0;
				}
			ditempvalb = 0;
			if ( ditempentity == dicarruec && dicarrscan > 1 && diPlayerUEC.GetCarrierStorage().GetItemCount( wViewDrone ) )
				ditempvalb = 14000;
			if ( dicarrscan > 1 && didrone )
				ditempvalb = 36000;
			if ( disatelliterefresh > 15 )
				ditempval = g_Game.QueryUnitsBySphere( "0 0 0", 100000, ditempobjects, 500 );
			else
				{
				ditempval = g_Game.QueryUnitsBySphere( ditempentity.GetPosition(), ditempvalb, ditempobjects, 500 );
				dicarrscancheck = false;
				}
			ditempval = 0;
			for ( ditempvale = 0; ditempvale < 500; ditempvale++ )
				{
				if ( ditempobjects[ditempvale] )
					{
					ditempval++;
					ditempobject = ditempobjects[ditempvale];
					ditempentity = ditempobject.GetEntity();
					ditemppos = dicurrent.GetPosition();
					diunitpos = ditempentity.GetPosition();
					ditemppos[2] = diunitpos[2];
					diobjectdist[ditempvale] = VectorLength( ditemppos - diunitpos ) + 1;
					}
				else
					ditempvale = 500;
				}
			for ( ditempvalb = 0; ditempvalb < 500; ditempvalb++ )
				{
				ditempvald = 1000000;
				ditempvalc = -1;
				for ( ditempvale = 0; ditempvale < ditempval; ditempvale++ )
					{
					if ( diobjectdist[ditempvale] && diobjectdist[ditempvale] <= ditempvald )
						{
						ditempvald = diobjectdist[ditempvale];
						ditempvalc = ditempvale;
						}
					}
				if ( ditempvalc > -1 )
					{
					ditempobjectsb[ditempvalb] = ditempobjects[ditempvalc];
					diobjectdistb[ditempvalb] = ditempvald;
					diobjectdist[ditempvalc] = 0;
					}
				else
					ditempvalb = 500;
				}
			ditempvald = 0;
			for ( ditempvale = 0; ditempvale < 500; ditempvale++ )
				{
				if ( diobjectdistb[ditempvale] )
					{
					ditempobject = ditempobjectsb[ditempvale];
					ditempentity = ditempobject.GetEntity();
					ditempstring = "";
					if ( IsInherited( ditempentity, Type("Carrier")) )
						ditempstring = "#radio_carrier";
					if ( IsInherited( ditempentity,Type("Walrus")) )
						{
						ditempstring = "#radio_walrus";
						ditemppos = ditempentity.GetPosition();
						if ( ditempentity.GetPlayer() != diPlayerUEC && ditempentity.WasDiscovered() && VectorLength( ditemppos - dicarrpos ) < 20000 )
							{
							if ( dinearestwalruspos == "0 0 0" || VectorLength( ditemppos - dicarrpos ) < VectorLength( dinearestwalruspos - dicarrpos ) )
								dinearestwalruspos = ditemppos;
							}
						}
					if ( IsInherited( ditempentity,Type("Manta")) && g_Game.SearchArray(dimantadummy, ditempentity) == -1)
						{
						if ( g_Game.SearchArray(diattackmanta, ditempentity) != -1 )
							ditempstring = "LONG RANGE ";
						ditempstring += "#radio_manta";
						}
					if ( IsInherited( ditempentity,Type("Turret")) && ditempentity.GetAssembly().GetEquipmentCount() )
						ditempstring = "- - -" + g_Game.ArticleName( ditempentity.GetAssembly().GetEquipment( 0 ) );
					if ( IsInherited( ditempentity,Type("Character")) )
						diaddsatelliteobject( ditempentity, "MapDroid", false, true );
					if ( IsInherited( ditempentity, Type("Building")) )
						{
						if ( di_enemy_unit_stealth && ditempentity.GetPlayer() != diPlayerUEC )
							ditempentity.LightColor = 0;
						build = ditempentity;
						ditempvalc = build.GetBuildingType();
						if ( ditempvalc == BuildingRadar )
							ditempstring = "#maptip_radar";
						if ( ditempvalc == BuildingCommandCenter )
							ditempstring = "#maptip_cc";
						if ( ditempvalc == BuildingUniPostBooster )
							ditempstring = "#maptip_sbooster";
						if ( ditempvalc == BuildingUniPostScrambler )
							ditempstring = "#maptip_scrambler";
						if ( ditempvalc == BuildingGarageManta )
							ditempstring = "#maptip_mgarage";
						if ( ditempvalc == BuildingGarageWalrus )
							ditempstring = "#maptip_wgarage";
						if ( ditempvalc == BuildingGenericPump && !direfueldiscovered && build.WasDiscovered() )
							direfueldiscovered = true;
						if ( ditempvalc == BuildingGenericHealth && !direpairdiscovered && build.WasDiscovered() )
							direpairdiscovered = true;
						if ( ditempvalc == BuildingGenericRearm && !direarmdiscovered && build.WasDiscovered() )
							direarmdiscovered = true;
						if ( ditempvalc == BuildingPier )
							{
							if ( diislstatus[dicurrentisland2] && VectorLength( dicarrpos - build.GetPosition() ) < 7000 )
								build.SetPlayer(diPlayerUEC);
							else
								build.SetPlayer(g_Game.GetPlayer(diislstatus[dicurrentisland2]));
							}
						}
					if ( ditempstring != "" && ditempentity.GetPlayer() != diPlayerUEC && ditempentity.VisibleOnMap() && disatelliterefresh < 16 )
						{
						ditempstringb = distring( ditempstring );
						if ( !dirussian )
							ditempstring = ditempstringb;
						ditacticaltextwidget = FindWidget ( ditacticalrootwidget[ditempvald], "TacticalScanText" );
						ditempsliderwidget = FindWidget ( ditacticalrootwidget[ditempvald], "Healthlevel" );
						ditemptextwidget = FindWidget ( ditacticalrootwidget[ditempvald], "Distance" );
						ditempvalc = ditempentity.GetHealth() * 100 * 1.35;
						ditempint = ditempvalc;
						ditempsliderwidget.SetSize ( ditempint, 10 );
						SetWidgetTextEx ( ditacticaltextwidget, 0, 0, ditempstring );
							ditacticaltextwidget.ClearFlags( WF_RALIGN );
						if ( disecondtoggle && strlen( ditempstringb ) > 18 )
							ditacticaltextwidget.SetFlags( WF_RALIGN );
						ditempvalc = diobjectdistb[ditempvale] / 32.1;
						ditempint = ditempvalc;
						if ( diobjectdistb[ditempvale] >= 32100 )
							SetWidgetTextEx ( ditemptextwidget, 0, 0, substr( ftoa ( diobjectdistb[ditempvale] / 32100 ) + ".0", 0, 3 ) + "km" );
						else
							SetWidgetTextEx ( ditemptextwidget, 0, 0, itoa ( ditempint ) + "m" );
						dicarrscanresult[ditempvald] = ditempentity;
						ditempvald++;
						if ( ditempvald >= dimaxresults )
							ditempvale = 500;
						}
					if ( ditempstring == "#radio_carrier" || ditempstring == "#radio_walrus" || ditempstring == "#radio_manta" || ditempstring == "LONG RANGE #radio_manta" )
						{
						if ( g_Game.SearchArray(disatelliteentity, ditempentity) == -1 && g_Game.SearchArray(dimantadummy, ditempentity) == -1 )
							{
							if ( g_Game.SearchArray(diunit, ditempentity) != -1 )
								diaddsatelliteobject( ditempentity, itoa( g_Game.SearchArray(diunit, ditempentity) + 1 ), false, true );
							else
								{
								if ( ditempstring == "#radio_carrier" )
									diaddsatelliteobject( ditempentity, "IconCarrier", false, true );
								if ( ditempstring == "#radio_walrus" )
									diaddsatelliteobject( ditempentity, "HudRadarWalrus", false );
								if ( ditempstring == "#radio_manta" )
									diaddsatelliteobject( ditempentity, "HudRadarManta", false );
								if ( ditempstring == "LONG RANGE #radio_manta" )
									diaddsatelliteobject( ditempentity, "HudMantaMedium", false );
								}
							}
						}
					if ( disatelliterefresh > 15 && ditempstring == "#radio_walrus" && !ditempentity.IsCarrierUnit() && ditempentity.GetAssembly().GetEquipmentCount() )
						{
						ditempvalb = g_Game.SearchArray(disatelliteentity, ditempentity);
						if ( ditempvalb != -1 )
							{
							if ( !ditempentity.HasPlan() )
								{
								diobj = ditempentity.GetAI();
								for ( ditempval = 0; ditempval < dispawnstationmax; ditempval++ )
									{
									build = dispawnstation[ditempval];
									if ( diobj && build && build.GetBuildingType() == BuildingGarageWalrus && VectorLength( ditempentity.GetPosition() - build.GetPosition() ) < 1500 )
										{
										diobjh = diobj.GetHandler();
										if ( diobjh )
											{
											diobjh.PlanAction(CreateMovementAction (Vector( dicarrpos[0], dicarrpos[1], 0)), false, true);
											ditempentity.AutoPilot(true);
											digarageclear[ditempvalb] = true;
											}
										}
									}
								}
							else
								{
								if ( digarageclear[ditempvalb] )
									{
									ditempentity.PlanClear();
									digarageclear[ditempvalb] = false;
									}
								}
							}
						}
					}
				else
					ditempvale = 500;
				}
			if ( disatelliterefresh < 16 || !di_tactical_scan_limit )
				{
				for ( ditempval = 0; ditempval < ditempvald; ditempval++ )
					{
					ditempvale = discreensizey / 2;
					ditempvalc = 13 * ( ditempvald / 2 );
					ditempvale -= ditempvalc;
					ditempvalb = 13 * ditempval;
					ShowWidget ( ditacticalrootwidget[ditempval], di_tactical_scan_limit );
					SetWidgetPos ( ditacticalrootwidget[ditempval], 5, ditempvale + ditempvalb, 150, 10 );
					}
				for ( ditempval = ditempvald; ditempval < 100; ditempval++ )
					{
					ShowWidget ( ditacticalrootwidget[ditempval], false );
					dicarrscanresult[ditempval] = NULL;
					}
				discanmaxdisplay = ditempvald;
				}
			if ( disatelliterefresh > 15 )
				{
				for ( ditempval = 0; ditempval < 5; ditempval++ )
					dicarrweaps[ditempval] = diPlayerUEC.GetCarrierStorage().GetItemCount ( dicarrweap[ditempval * 3 + 1] ) + diPlayerUEC.GetCarrierStorage().GetItemCount ( dicarrweap[ditempval * 3 + 2] ) + diPlayerUEC.GetCarrierStorage().GetItemCount ( dicarrweap[ditempval * 3 + 3] )
				disatelliterefresh = 0;
				}
			}
		}
	didroidpath++;
	if ( didroidlaunchcounter )
		didroidlaunchcounter--;
	for ( ditempvalb = 5; ditempvalb < 9; ditempvalb++ )
		{
		didroidsinrange[ditempvalb] = 0;
		didroidsactive[ditempvalb] = 0;
		}
	for ( ditempval = 1; ditempval < 17; ditempval++ )
		{
		if ( didroidleader[ditempval] )
			{
			ditempentity = didroidleader[ditempval];
			if ( g_Game.SearchArray(diunit, ditempentity) == -1 && g_Game.SearchArray(didroid, ditempentity) == -1 || ditempentity.GetHealth() == 0 )
				didroidleader[ditempval] = NULL;
			}
		}
	for ( ditempval = 1; ditempval < 33; ditempval++ )
		{
		ditempentity = didroid[ditempval];
		diunitpos = didroidfpos[ditempval];
		if ( ditempentity && diunitpos != "0 0 0" )
			{
			ditempentity.DisableComplex();
			ditempentity.SetPosition(didroidfpos[ditempval]);
			ditempentity.EnableComplex();
			if ( ditempentity == dicurrent && dicarruec )
				{
				dicarruec.SelectForPlayer();
				ditempentity.SelectForPlayer();
				}
			}
		}
	for ( ditempval = 1; ditempval < 17; ditempval++ )
		{
		if ( didroidpathb[ditempval] )
			didroidpathb[ditempval] = didroidpathb[ditempval] - 1;
		diunit2 = diunit[ditempval - 1 / 4 + 5];
		ditempentity = didroid[ditempval];
		if ( ditempentity )
			{
			didroidsactive[ditempval - 1 / 4 + 5] = didroidsactive[ditempval - 1 / 4 + 5] + 1;
			if ( didroidlaunchcounter )
				{
				ditempentity.ForcePlayer(diPlayerUEC);
				if ( ditempentity.IsComplexlyDisabled() )
					ditempentity.EnableComplex();
				ditempentity.AutoPilot(true);
				}
			ditempvale = 0;
			for ( ditempvalb = 1; ditempvalb < 9; ditempvalb++ )
				{
				diunit1 = diunit[ditempvalb];
				if ( diunit1 && dihasplan[ditempvalb] && !diisplansuspended[ditempvalb] && VectorLength( diunit1.GetPosition() - ditempentity.GetPosition() ) <= 500 || diunit1 && !dienemyinrange[ditempvalb] && VectorLength( diunit1.GetPosition() - ditempentity.GetPosition() ) <= 1000 )
					ditempvale = ditempvalb;
				}
			if ( ditempvale )
				ditempentity.SetInvincibility(true);
			else
				ditempentity.SetInvincibility(false);
			if ( diunit2 && VectorLength( diunit2.GetPosition() - ditempentity.GetPosition() ) <= 650 )
				{
				didroidsinrange[ditempval - 1 / 4 + 5] = didroidsinrange[ditempval - 1 / 4 + 5] + 1;
				if ( ditempentity.GetHealth() > 0 )
					ditempentity.Heal( 0.002 );
				if ( didroidpath == 2 && ditempentity.GetAmmo() < 400 )
					ditempentity.AddAmmo( 1 );
				}
			diunit1 = didroidleader[ditempval];
			if ( !didroidpathb[ditempval] && diunit1 && ditempentity != dicurrent && ditempentity != didroidleader[ditempval] || didroidrecall[ditempval] && diunit1 && ditempentity != dicurrent )
				{
				ditempentity.EnableAI();
				diobj = ditempentity.GetAI();
				if ( diobj )
					{
					diobjh = diobj.GetHandler();
					GetMatrix4(diunit1, dirotvector);
					diunitpos = VectorMatrixMultiply4(dirotvector, didroidpos[ditempval]);
					ditempvale = GetTerrainHeight(diunitpos);
					diunitpos[2] = ditempvale;
					if ( diunitpos[2] > 10 && VectorLength( ditempentity.GetPosition() - diunit1.GetPosition() ) > 600 || didroidrecall[ditempval] && diunitpos[2] > 10 )
						{
						if ( !didroidpathb[ditempval] )
							{
							ditempentity.PlanClear();
							didroidpathb[ditempval] = 30;
							}
						diobjh.PlanAction(CreateMovementAction (diunitpos), false, false);
						}
					if ( ditempentity.GetAmmo() == 0 || VectorLength( diunit1.GetPosition() - ditempentity.GetPosition() ) > 2000 || didroidrecall[ditempval] )
						diobj.SetStance( STANCE_PASSIVE );
					if ( VectorLength( diunit1.GetPosition() - ditempentity.GetPosition() ) <= 600 )
						{
						if ( ditempentity.GetAmmo() >= 300 || ditempentity.GetAmmo() > 0 && !diunit1.GetAssembly().IsWalrus() )
							{
							if ( didroidrecall[ditempval] )
								didroidrecall[ditempval] = didroidrecall[ditempval] - 1;
							if ( !didroidrecall[ditempval] )
								diobj.SetStance( STANCE_AGGRESSIVE );
							}
						}
					ditempentity.AutoPilot(true);
					ditempentity.PlanEnable();
					}
			else
				g_Game.PlayGameSound( "fx/beep_locking" );
				}
			if ( ditempentity.GetHealth() == 0 )
				{
				ditempentity.SetPosition( "5000 5000 -5000" );
				ditempentity.Destroy();
				if ( diunit2 && ditempentity == dicurrent )
					diunit2.SelectForPlayer();
				didroid[ditempval] = NULL;
				didroidleader[ditempval] = NULL;
				}
			}
		}
	for ( ditempval = 1; ditempval < 33; ditempval++ )
		{
		didroidfpos[ditempval] = "0 0 0";
		diunit2 = NULL;
		if ( ditempval < 17 )
			diunit2 = diunit[ditempval - 1 / 4 + 5];
		ditempentity = didroid[ditempval];
		if ( !diunit2 && ditempentity || diunit2 && ditempentity && diisdocked[ditempval - 1 / 4 + 5] )
			{
			for ( ditempvalb = 1; ditempvalb < 5; ditempvalb++ )
				{
				ditempvald = diPlayerUEC.GetCarrierStorage().GetItemCount(13457 + ditempvalb + 4) + didroidsactive[ditempvalb + 4];
				diunit1 = diunit[ditempvalb + 4];
				if ( diunit1 && ditempvald < 4 && VectorLength(diunit1.GetPosition() - ditempentity.GetPosition()) <= 1200 )
					{
					ditempvalc = ditempvalb * 4 - 3;
					for( ditempvald = ditempvalc; ditempvald < ditempvalc + 4 ; ditempvald++ )
						{
						if ( !didroid[ditempvald] )
							{
							didroid[ditempvald] = didroid[ditempval];
							didroidleader[ditempvald] = diunit1;
							didroid[ditempval] = NULL;
							didroidsactive[ditempvalb + 4] = didroidsactive[ditempvalb + 4] + 1;
							}
						}
					}
				}
			}
		}
	if ( dicurrent && dicarruec && !direset && !diresetprev && dicurrent.GetAssembly().IsWalrus() != dicurrent.GetAssembly().IsManta() )
		{
		if ( !digm && !dicurrentunit || dicurrentisland2 && !dicurrentunit )
			dicurrent.SetFuel ( 1.0 );
		if ( KeyState(KC_LCONTROL) && diinmap && dipointedentity && dipointedentity != dicurrent )
			{
			for ( ditempval = 0; ditempval < 9; ditempval++ )
				{
				ditempentity = dicurrent;
				if ( ditempval )
					ditempentity = diunit[ditempval];
				if ( ditempentity )
					{
					if ( ditempentity == dicurrent || ditempentity != dicurrent && ditempentity.IsWingman() )
						{
						ditempentity.PlanClear();
						if ( ditempentity.IsDocked() )
							ditempentity.GetAssembly().RequestDeploy();
						diobj = ditempentity.GetAI();
						if ( diobj )
							{
							diobjh = diobj.GetHandler();
							if ( ditempentity.GetAssembly().IsWalrus() )
								ditemppos = FindClosestReachable(dipointedentity.GetPosition(), 1000, diobj);
							else
								ditemppos = dipointedentity.GetPosition();
							diobjh.PlanAction(CreateMovementAction (ditemppos), false, false);
							ditempentity.AutoPilot(true);
							ditempentity.PlanEnable();
							}
						}
					}
				}
			}
		if ( !dicurrent.IsCarrierUnit() && VectorLength (dicurrent.GetPosition() - dicarruec.GetPosition() ) >= dicarruec.GetCarrierTelemetry() )
			{
			if ( !g_GM || diloadedisland != "Thermopylae" && dicurrent.GetName() != "MyrikManta" )
				{
				dicurrent.ProcessDock();
				dicurrent.PlanClear();
				dicurrent = NULL;
				}
			}
		}
	if ( didrone && !didocking[0] && dicurrent && dicurrent.IsActorOnBoard() || didrone && !didocking[0] && dicurrent && strstr (dicurrent.GetName(), "Myrik" ) != -1 )
		{
		diunitpos = didrone.GetPosition();
		ditemppos = dicurrent.GetPosition();
		diunitpos[2] = ditemppos[2];
		diobj = didrone.GetAI();
		if ( diobj && VectorLength( diunitpos - ditemppos ) > 4000 )
			{
			diobjh = diobj.GetHandler();
			diobjh.PlanAction(CreateIMovementAction (dicurrent.GetPosition()) , false, true);
			didrone.PlanEnable();
			}
		else
			didrone.PlanClear();
		if ( dicarruec && VectorLength (dicarruec.GetPosition() - didrone.GetPosition()) > 450000 * didrone.GetFuel() )
			{
			didrone.PlanClear();
			didocking[0] = true;
			}
		}
	diforcedlodtimer++;
	if ( diforcedlod != di_force_lod || di_force_lod && diforcedlodtimer > 50 )
		{
		diforcedlodtimer = 0;
		diforcedlod = di_force_lod;
		ditempentity = GetNextEntity(NULL, NULL);
		while (ditempentity)
			{
			SetFixedLOD( ditempentity, diforcedlod - 1 );
			ditempentity = GetNextEntity(ditempentity, NULL);
			}
		}
	diexecutions++;
	if ( didroidpath >= 2 )
		didroidpath = 0;
	}
		return MyPlayer.GetStockpile();
	}

	int StockpileID()
	{
		return Stockpile().GetArea();
	}

	IslandModel AtIsland()
	{
		return MyPlayer.GetCurrentIsland();
	}

	int AtIslandID()
	{
		return AtIsland().GetArea();
	}


	Carrier MyCarrierEntity()
	{
		return MyPlayer.GetCarrier();
	}

	EntityBasic MyCarrier()
	{
		return MyPlayer.GetCarrierBasic();
	}

	bool HasMyCarrierAllDocked()
	{
		EntityBasic assemblies[16];
		int numAssemblies = MyPlayer.GetAssemblies( assemblies );
		for( int i = 0; i < numAssemblies; i++ )
		{
			EntityBasic assembly = assemblies[i];
			if( !assembly.IsDestroyed() && assembly.IsActive() )
				return false;
		}

		return true;
	}

	bool AtStockpile()
	{
		if( AtIsland() == Stockpile() ) return true;
		return false;
	}

	bool Docked()
	{
		return MyCarrier().IsDocked(NULL);
	}

	string GetName()
	{
		return "Default expert";
	}

	//@}
};
/**@}*/

/** \addtogroup AIAPI */
/**@{*/
class ScriptedFightGroup extends FightGroup
{
	Object Members[32];
	int MemberCount;

	//! helper: add member
	void AddMember( Object member )
	{
		Members[MemberCount] = member;
		MemberCount++;
	}

	//! helper: remove member
	void RemoveMember( Object member )
	{
		for( int i = 0; i < 32; i++ )
		{
			Object mem = Members[i];
			if( !mem )
			{
				Print( "WARNING: FightGroup::RemoveMember failed!" );
				return;
			}

			if( mem == member )
			{
				Members[i] = NULL;
				--MemberCount;
				break;
			}
		}

		//		CompactPtrArray( Members, 32 );
		//hmm :(
		int freeIndex = -1;
		for( i = 0; i < 32; i++ )
		{
			Object c = Members[i];
			if( !c )
			{
				// empty slot
				if( freeIndex < 0 ) freeIndex = i;
			}
			else
			{
				if( freeIndex >= 0 )
				{
					Members[freeIndex] = c;
					Members[i] = NULL;

					// select next free index
					for( int j = freeIndex + 1; j < 32; j++ )
					{
						c = Members[j];
						if( !c )
						{
							freeIndex = j;
							break;
						}
					}
				}
			}
		}
	}

	//! new member event
	void OnMemberAdded( Object member )
	{
	}

	//! member removed event
	void OnMemberRemoved( Object member )
	{
	}

	//! process logic
	void Process()
	{
	}
};


//-----------------------------------------------------------------------------
// ADDITIONAL API (should be in [AILib.h] but requires some classes defined here)
//-----------------------------------------------------------------------------


//! info about tactic-managed object
class TacticObjInfo
{
	ObjectHandlerBase Handler;
	ScriptedTactic AssignedTactic;
};

//! info about tactic's Point-of-Interest
class TacticPOIInfo
{
	Object POI;
	ScriptedTactic AssignedTactic;
};
/**@}*/